#include <stdio.h>
#include <math.h>

#include <iostream>
#include <algorithm>

using std::max ;
using std::min ;
using std::cout ;
using std::endl ;

// DIM is the maximum dimension of the temperature array
const int DIM=1024;
// Allocate an array of temperatures
double T[DIM][DIM] ;


/*******************************************************************/
/* A routine for writing out our Two Dimensional Array as a pixmap */
/* *****************************************************************/
/* Output a tile as a portable pixmap (ppm) file                   */
/* You don't need to understand this code                          */
/*******************************************************************/

void tile2ppm(double tile[DIM][DIM], int dim,char *filename) {
  FILE *fp ;

  if((fp = fopen(filename,"w")) == NULL) {
      perror(filename) ;
      exit(-1) ;
  }
  int idim = dim ;
  int jdim = dim ;
  unsigned char R[256],G[256],B[256] ;
  int i,j ;
   
  /* Create Color Map */
  for(i=0;i<256;++i) {
    double s = (double)(i)/255.0 ;
    double t = (1.-s) ;
    double th = s*3.1415927*.5 ;
    double th2 = -3.1415927+s*3.1415927*2 ;
    
    R[i] = min(int(800*s),255) ;
    G[i] = min(int(256*s*s),255) ;
    B[i] = max(min(50+int(512*(0.5-s)),255),
               min(int(512*(s-0.5)),255)) ;
  }
                                                                                
  /* Write ppm file header */
  fprintf(fp,"P6\n%d %d\n255\n",idim,jdim) ;
 
  /* Write Image data */
  for(i=0;i<idim;++i)
    for(j=0;j<jdim;++j) {
      int val = int(tile[i][j]*255) ; ;
      fputc(R[val],fp) ;
      fputc(G[val],fp) ;
      fputc(B[val],fp) ;
    }
  fclose(fp) ;
}

main(int ac, char *av[]) {

  // read in command line arguments for size of simulation and
  // convergence tolerance on the iterative convergence
  int dim = DIM ;
  if(ac >= 2) {
    dim = atoi(av[1]) ;
    if(dim>DIM)
      dim = DIM ;
  }
  double tol = 1e-6;
  if(ac >= 3) {
    tol = atof(av[2]) ;
  }

  // Initialize temperature to zero
  for(int j=0;j<dim;++j) {
    for(int i=0;i<dim;++i) {
      T[i][j] = 0 ;
    }
  }

  // Set up boundary conditions for hot spot on left and 5 small hot spots
  // on top.  
  for(int i=dim/10;i<dim-dim/4;++i) {
    T[dim-i-1][0] = 1 ;
  }
  int dd = dim/10;
  for(int i=0;i<dim;++i) 
    if((i/dd)&1)
      T[0][i] = 1 ;


  // We will iterate until rms error drops below some threshold
  double rms_error ;
  double error_sum = 0.0 ;
  
  // Successive Overrelaxation Parameter (To improve iteration performance)
  const double chi=1.85 ;
  // Count the number of times we iterate to achive convergence
  int count = 0 ;

  // Begin looping until error tolerance is met
  do {
    count++ ;

    // Red Black Gauss Seidel, Update red squares then black squares (as
    // in checker board pattern.  Note, each iterate is independent in
    // the red-black Gauss Seidel algorithm

    // here we do the red squares
    for(int i=1;i<dim-1;i+=1)
      for(int j=1+(i&1);j<dim-1;j+=2) 
        T[i][j] = T[i][j]*(1-chi)+
                  chi*0.25*(T[i-1][j]+T[i+1][j]+T[i][j-1]+T[i][j+1]) ;

    // now the black squares
    for(int i=1;i<dim-1;i+=1) 
      for(int j=2-(i&1);j<dim-1;j+=2) 
        T[i][j] = T[i][j]*(1-chi)+
                  chi*0.25*(T[i-1][j]+T[i+1][j]+T[i][j-1]+T[i][j+1]) ;

    // Check root mean square error:  If the solution is converged then
    // each point's temperature should equal the average of the neighbors.
    // The extent that this isn't satisfied is the local error.  The global
    // error is found by summing the squares of the local errors.
    error_sum = 0.0 ;
    for(int j=1;j<dim-1;++j) 
      for(int i=1;i<dim-1;++i) {
        double  err =(T[i-1][j]+T[i+1][j]+T[i][j-1]+T[i][j+1]) - 4.*T[i][j] ;
        error_sum += err*err ;
      }
    rms_error = sqrt(error_sum/(dim*dim)) ;
    cout << "rms_error = " << rms_error << endl ;
  } while(rms_error > tol) ;

  cout << "num_iters to convergence = " << count << endl ;


  /* We've finished processing so write out a pixmap file so we can see the
     results! */
  
  char buf[512] ;
  sprintf(buf,"%s_%d.ppm",av[0],dim) ;
  tile2ppm(T,dim,buf) ;
}






    
